{	This program is copyright (c) 1993 - 2005 Jos Kunst, the Jos Kunst heirs.	}
{	This program is free software. You can redistribute it and/or modify it 	}
{	under the terms of version 2 of the GNU General Public License, which you	}
{	should have received with it.												}
{	This program is distributed in the hope that it will be useful, but			}
{	without any warranty, expressed or implied.									}

unit workspace;
{relatie met players: in players staat de 'objectieve' constructie van instrumenten en spelers beschreven, }
{in workspace zoveel mogelijk van het voor doposition en productions nodige werk ermee}
interface
	uses
		translations, chordscanner, chordworker, randomizers, outputs, players;

	const
		minlength = 3;
		maxfactor = 1.6;
	var
		maxlength: byte;

	procedure resolvedifficulties (fixedtones: pitchset;
									var mobiletones: pitchset;
									var leftovers: pitchset;
									instr: instruments;
									maxcard: byte);
{given a set of fixed tones and a set that may be transposed through octaves,}
{find a way to combine them into a chord that is playable for a given string instrument }
{in a given circumstances limiting the cardinality of the resulting chord}
	function simplified (dr, frompos: byte): byte;
{given a duration and its starting point, replaces the duration by another }
{that has an end point whose precision in the measure is related to its length}
	procedure dodoubledicing (target, precision: byte;
									var result: byte);
{given two targets and two corresponding precisions, performs }
{both random operations and stores their results in one doubledigit byte }
	procedure pullbackchord (var cs: pitchset;
									anchor: byte;
									var result: integer);
{given a chord C and a tone T, moves C so that T becomes one of its tones}
	procedure updateanalysis (cs: pitchset;
									resulstep: integer);
{given a chord and an interval over which it has been moved, calculates new dsg items}
	procedure givitapart (i: instruments;
									cs: pitchset;
									var p: byte);
{strikes a bargain between good instrumental register and a given chord's requirements }
{ in assigning one of its tones to a given instrument}

implementation

	procedure resolvedifficulties (fixedtones: pitchset;
									var mobiletones: pitchset;
									var leftovers: pitchset;
									instr: instruments;
									maxcard: byte);
		var
			o, p, q, r, s, diff1, handpos1, diff2, handpos2, lowest, low, high, highest: byte;
			mobtones, octaves: pitchset;
	begin
		leftovers := [];
		if card(fixedtones + mobiletones) > 1 then
			begin
				mobiletones := mobiletones - fixedtones; {--- make the two sets disjoint ---}
				if card(fixedtones) > maxcard then
					begin
						writeln('too many fixed tones for ', name(instr), '!');
						exit(resolvedifficulties)
					end
				else
					while card(fixedtones + mobiletones) > maxcard do
						begin
							r := randnum(card(fixedtones + mobiletones));
							p := bottomnote - 1;
							repeat
								p := p + 1;
								if p in (fixedtones + mobiletones) then
									r := r - 1
							until (r = 0) or (p = topnote);
							if p in mobiletones then
								begin
{    writeln('to leftovers: ', nameof(p)); --> for testdifficulties}
									leftovers := leftovers + [p];
									mobiletones := mobiletones - [p]
								end
						end;
				checkdifficulty((fixedtones + mobiletones), instr, diff1, handpos1, diff2, handpos2);
{    writeln('difficulty ', diff1 : 1, ' (', base10to2(diff1), ') in LHpos ', posvalue(2, handpos1, 16) : 1, ' plus ', posvalue(1, handpos1, 16) : 1, ' for ', name(instr), '...'); --> for testdifficulties}
				if diff1 = 255 then
					begin
						givelimitsof(instr, lowest, low, high, highest);
						p := lowest;
{--------------------------------- even... --------------------------------}
						p := bottomnote;
{--------------------------------- even... --------------------------------}
						mobtones := []; {---the intended replacement set ----}
						repeat
							if p in mobiletones then
								begin
{writeln('... working on ', nameof(p), '...'); --> for testdifficulties}
{--- providing for a stop condition and jettisoning the intractable cases---}
									mobiletones := mobiletones - [p];
 {--------- first see whether adding just this one tone is possible -------}
									mobtones := mobtones + [p];
									checkdifficulty((fixedtones + mobtones), instr, diff1, handpos1, diff2, handpos2);
{writeln('difficulty ', diff1 : 1, ' (', base10to2(diff1), ') in LHpos ', posvalue(2, handpos1, 16) : 1, ' plus ', posvalue(1, handpos1, 16) : 1, '...'); --> for testdifficulties}
{----- if not, try its octave transpositions, close ones first the---------}
									if diff1 = 255 then
										begin
{write(nameof(p), ': impossible...'); --> for testdifficulties}
{ ---- determine the transpositions; first 1 octave, then 2, then 3 ----}
											o := 1;
											repeat
												octaves := [];
												q := p;
												if (q - o * 12 >= lowest) and (q - o * 12 <= highest) then
													octaves := octaves + [q - o * 12];
												if (q + o * 12 <= highest) and (q + o * 12 >= lowest) then
													octaves := octaves + [q + o * 12];
{------ choose each one in turn to try replacing the problem tone, thereby emptying the octave set ----- }
												repeat
													mobtones := mobtones - [q];
													r := randnum(2);
													if r = 1 then
														begin
															q := lowest - 1;
															repeat
																q := q + 1
															until (q in octaves) or (q > highest)
														end
													else
														begin
															q := highest + 1;
															repeat
																q := q - 1
															until (q in octaves) or (q < lowest)
														end;
													if q in octaves then
														begin
															octaves := octaves - [q];
															mobtones := mobtones + [q];
{writeln(' - will octave variant ', nameof(q), ' work?...'); --> for testdifficulties}
															checkdifficulty((fixedtones + mobtones), instr, diff1, handpos1, diff2, handpos2);
{writeln('difficulty ', diff1 : 1, ' (', base10to2(diff1), ') in LHpos ', posvalue(2, handpos1, 16) : 1, ' plus ', posvalue(1, handpos1, 16) : 1, '...'); --> for testdifficulties}
															if diff1 = 255 then
																begin
{write('no; so ditch ', nameof(q), '! '); --> for testdifficulties}
																	mobtones := mobtones - [q]
																end
{else writeln('octave ', nameof(q), ' OK...') --> for testdifficulties}
														end
												until (diff1 < 255) or (octaves = []);
												o := o + 1
											until (diff1 < 255) or (o = 4);
											if diff1 = 255 then
												begin
													leftovers := leftovers + [p];
{writeln(nameof(p), ' --> to leftovers...') --> for testdifficulties}
												end
										end
{else    writeln(nameof(p), ' OK...'); --> for testdifficulties}
								end;
							p := p + 1
						until (mobiletones = []) or (p > topnote);
						checkdifficulty((fixedtones + mobtones), instr, diff1, handpos1, diff2, handpos2);
{    writeln('difficulty ', diff1 : 1, ' (', base10to2(diff1), ') in LHpos ', posvalue(2, handpos1, 16) : 1, ' plus ', posvalue(1, handpos1, 16) : 1, '...'); --> for testdifficulties}
						mobiletones := mobtones
					end
			end
		else
			begin
				p := bottomnote - 1;
				repeat
					p := p + 1
				until (p in mobiletones) or (p > topnote);
				if p in mobiletones then
					begin
						givelimitsof(instr, lowest, low, high, highest);
						while p > highest do
							p := p - 12;
						while p < lowest do
							p := p + 12;
{    writeln('this single note set contains a playable ', nameof(p), ' for ', name(instr));}
						mobiletones := [p]
					end
			end
	end;

	function simplified (dr, frompos: byte): byte;
		var
			taildr: byte;
{ -----------time is measured in triplet ninetysixths!--------}
		procedure paredownfromhead;
			var
				headdr: byte;
		begin
			taildr := dr;
			case frompos of
				0: 
					headdr := 0;
				1: 
					if dr < 24 then
						headdr := 10
					else
						headdr := 22;
				2: 
					if dr < 24 then
						headdr := 9
					else
						headdr := 21;
				3: 
					if dr < 24 then
						headdr := 8
					else
						headdr := 20;
				4: 
					if dr < 24 then
						headdr := 6
					else
						headdr := 18;
				5: 
					if dr < 24 then
						headdr := 4
					else
						headdr := 16;
				6: 
					if dr < 24 then
						headdr := 3
					else
						headdr := 15;
				7: 
					if dr < 24 then
						headdr := 2
					else
						headdr := 14;
				8: 
					if dr < 24 then
						headdr := 0
					else
						headdr := 12;
				9: 
					headdr := 10;
				10: 
					headdr := 9;
				11: 
					headdr := 8;
				12: 
					headdr := 6;
				13: 
					headdr := 4;
				14: 
					headdr := 3;
				15: 
					headdr := 2
			end;
			if taildr > headdr then
				taildr := taildr - headdr;
			if dr >= 24 then
				while taildr > 24 do
					taildr := taildr - 24
			else
				while taildr > 12 do
					taildr := taildr - 12
		end;
	begin
		if dr = 255 then
			simplified := 255
		else if dr >= 128 then
			begin
				paredownfromhead;
{alleen kwarten toegelaten}
				simplified := dr + (24 - taildr);
			end
		else if dr >= 64 then
			begin
				paredownfromhead;
{alleen achtsten}
				if taildr > 12 then
					simplified := dr + (24 - taildr)
				else
					simplified := dr + (12 - taildr)
			end
		else if dr >= 32 then
			begin
				paredownfromhead;
{achtsten en trioolachtsten}
				if taildr > 16 then
					simplified := dr + (24 - taildr)
				else if taildr > 12 then
					simplified := dr + (16 - taildr)
				else if taildr > 8 then
					simplified := dr + (12 - taildr)
				else
					simplified := dr + (8 - taildr)
			end
		else if dr >= 24 then
			begin
				paredownfromhead;
{trioolachtsten en zestienden}
				if taildr > 18 then
					simplified := dr + (24 - taildr)
				else if taildr > 16 then
					simplified := dr + (18 - taildr)
				else if taildr > 12 then
					simplified := dr + (16 - taildr)
				else if taildr > 8 then
					simplified := dr + (12 - taildr)
				else if taildr > 6 then
					simplified := dr + (8 - taildr)
				else
					simplified := dr + (6 - taildr)
			end
		else if dr >= 16 then
			begin
				paredownfromhead;
{zestienden en trioolzestienden}
				if taildr > 8 then
					simplified := dr + (12 - taildr)
				else if taildr > 6 then
					simplified := dr + (8 - taildr)
				else if taildr > 4 then
					simplified := dr + (6 - taildr)
				else
					simplified := dr + (4 - taildr)
			end
		else if dr >= 12 then
			begin
				paredownfromhead;
{trioolzestienden en tweendertigsten}
				if taildr > 9 then
					simplified := dr + (12 - taildr)
				else if taildr > 8 then
					simplified := dr + (9 - taildr)
				else if taildr > 6 then
					simplified := dr + (8 - taildr)
				else if taildr > 4 then
					simplified := dr + (6 - taildr)
				else if taildr > 3 then
					simplified := dr + (4 - taildr)
				else
					simplified := dr + (3 - taildr)
			end
		else
			begin
				if frompos mod 4 = 0 then
					case dr of
						1, 5, 7, 11: 
							simplified := dr + 1;
						otherwise
							simplified := dr
					end
				else if frompos mod 4 = 1 then
					case dr of
						1, 3, 5, 9, 11: 
							simplified := dr + 1;
						otherwise
							simplified := dr
					end
				else if frompos mod 4 = 2 then
					case dr of
						1: 
							simplified := 3;
						2, 4, 8, 10: 
							simplified := dr + 1;
						otherwise
							simplified := dr
					end
				else if frompos mod 4 = 3 then
					case dr of
						1, 3, 7, 9: 
							simplified := dr + 1;
						otherwise
							simplified := dr
					end
			end
	end;

	procedure dodoubledicing (target, precision: byte;
									var result: byte);
		var
			p, q, r, s: byte;
	begin
		p := precision div 10;
		q := precision mod 10;
		case p of
			0: 
				r := randnum(7);
			1: 
				r := l1dce(7, target div 10);
			2: 
				r := l2dce(7, target div 10);
			3: 
				r := l3dce(7, target div 10);
			4: 
				r := l4dce(7, target div 10);
			5: 
				r := l5dce(7, target div 10);
			6: 
				r := target div 10;
			otherwise
				writeln('illegal precision factor!')
		end;
		case q of
			0: 
				s := randnum(7);
			1: 
				s := l1dce(7, target mod 10);
			2: 
				s := l2dce(7, target mod 10);
			3: 
				s := l3dce(7, target mod 10);
			4: 
				s := l4dce(7, target mod 10);
			5: 
				s := l5dce(7, target mod 10);
			6: 
				s := target mod 10;
			otherwise
				writeln('illegal precision factor!')
		end;
		result := 10 * r + s
	end;

	procedure pullbackchord (var cs: pitchset;
									anchor: byte;
									var result: integer);
		var
			c, closest, dist: byte;
			cr: chordrec;
{zzzzzz zou moeten checken of anchor geen 2 is, bijvoorbeeld: '-'  ... }
{en er dan iets over moeten zeggen, of er iets aan doen, bijvoorbeeld: anchor vervangen door nearestoctaveof(anchor)}
	begin
		dist := 200;
		for c := bottomnote to topnote do
			if c in cs then
				if abs(c - anchor) < dist then
					begin
						dist := abs(c - anchor);
						closest := c
					end;
		if anchor < closest then
			for c := bottomnote to topnote do
				if c in cs then
					begin
						cs := cs - [c];
						cs := cs + [c - dist]
					end;
		if anchor > closest then
			for c := topnote downto bottomnote do
				if c in cs then
					begin
						cs := cs - [c];
						cs := cs + [c + dist]
					end;
		if anchor > closest then
			result := dist
		else
			result := -dist
	end;

	procedure updateanalysis (cs: pitchset;
									resulstep: integer);
		var
			p: byte;
	begin
		hroot := hroot + resulstep;
		p := 1;
		repeat
			carr[p, 1] := carr[p, 1] + resulstep;
			p := p + 1
		until (carr[p, 1] = 0) or (p = 102)
	end;

	procedure givitapart (i: instruments;
									cs: pitchset;
									var p: byte);
		var
			loest, lo, hi, hiest: byte;

		procedure givitapitch (ins: instruments;
										var pit: byte);
			function weiiightedchoice (lowest, low, high, highest: byte): byte;
				var
					throw: byte;
			begin
				throw := randnum((3 * (high - low)) + (highest - lowest)) + (lowest - 1);
				if (throw <= highest) then
					weiiightedchoice := throw
				else
					begin
						while throw > highest do
							throw := throw - (high - low);
						throw := throw - (highest - high);
						weiiightedchoice := throw;
					end;
			end;
		begin
			pit := weiiightedchoice(loest, lo, hi, hiest);
		end;

		procedure getintochord (var pi: byte;
										ch: pitchset);
			var
				pplus, pminus, lolimit, hilimit: byte;
		begin
			lolimit := lo;
			hilimit := hi;
			pplus := pi;

			while not ((pplus in ch) or (pplus > hilimit)) do
				pplus := pplus + 1;
			if (pplus > hilimit) or (pplus < lolimit) then
				pplus := 0;
			pminus := pi;
			while not ((pminus in ch) or (pminus < lolimit)) do
				pminus := pminus - 1;
			if (pminus < lolimit) or (pminus > hilimit) then
				pminus := 0;
			if abs(pplus - pi) <= abs(pminus - pi) then
				pi := pplus
			else
				pi := pminus
		end;  {getintochord}

	begin {givitapart}
		givelimitsof(i, loest, lo, hi, hiest);
		givitapitch(i, p);
		getintochord(p, cs)
	end;{givitapart}

end.